#!/bin/sh
# Author: Alice

GRIDLOK="./gridlok.sh"

ulimit -t 30

logfile=test.log
if [ -f $logfile ]
then
    rm -f $logfile
fi
if [ -d tests/output ]
then
    rm -rf tests/output
fi
mkdir tests/output

if [ $# -ge 1 ]
then
    files=$@
else
    files=$(ls tests/test_*.gl tests/fail_*.gl)
fi

for file in $files
do
    case $file in
        *test_*)
            filename=$(echo $file | sed 's/.gl//' | sed 's/tests\///')
            $GRIDLOK $file
            ./test > tests/output/${filename}.out
            if [ -f tests/${filename}.out ]
            then
                diff -b tests/${filename}.out tests/output/${filename}.out > tests/output/${filename}.diff
                if [ -s tests/output/${filename}.diff ]
                then
                    echo "${filename}: FAILED" | tee -a $logfile
                else
                    rm tests/output/${filename}.diff
                    echo "${filename}: OK" | tee -a $logfile
                fi
            else
                echo "${filename}: FAILED - tests/${filename}.out does not exist" | tee -a $logfile
            fi
            ;;
        *fail_*)
            filename=$(echo $file | sed 's/.gl//' | sed 's/tests\///')
            $GRIDLOK $file 2> tests/output/${filename}.out
            if [ -f tests/${filename}.out ]
            then
                diff -b tests/${filename}.out tests/output/${filename}.out > tests/output/${filename}.diff
                if [ -s tests/output/${filename}.diff ]
                then
                    echo "${filename}: FAILED" | tee -a $logfile
                else
                    rm tests/output/${filename}.diff
                    echo "${filename}: OK" | tee -a $logfile
                fi
            else
                echo "${filename}: FAILED - tests/${filename}.out does not exist" | tee -a $logfile
            fi
            ;;
        *)
            echo "unknown file type $file"
            exit 1
            ;;
    esac
done

rm tests/output/*.out
rm test
