(* Author: Laura *)
open Printf
open Ast
open Sast

let file = "test.c"

let string_of_op = function
  Add -> "+"
  | Sub -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Mod -> "%"
  | StrEqual -> failwith("Wrong way to get stringeql!")
  | Equal -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq ->  ">="
  | And -> "&&"
  | Or -> "||"

let string_of_uop = function
  | Not -> "!"


let string_of_typ = function
    Int -> "int"
  | Bool -> "int"
  | Str -> "char *"
  | Piece | Void -> failwith("Semantic check failed to eliminate the tpying of piece and voids. ")

let rec string_of_expr = function
  | Sast.Parenth(e) ->
    let _,e = e in "(" ^ string_of_expr e ^ ")"
  | Sast.Id (i) -> if (i="turn" || i="x"||i="y") then i
      else if i="height" then "HEIGHT"
      else if i="width" then "WIDTH"
      else i^"UDV" 
  | Sast.BoolLiteral(b) -> if b=true then string_of_int 1 else string_of_int 0
  | Sast.IntLiteral(i) -> string_of_int i
  | Sast.StringLiteral(s) -> s
  | Sast.Binop(e1,o,e2) -> let _,e1=e1 and _,e2=e2 and t,_=e1 in
    if o=Ast.StrEqual then ("strcmp(" ^ string_of_expr e1 ^ "," ^ string_of_expr e2 ^ ")==0")
    else if t=Ast.Str then ("strcmp(" ^ string_of_expr e1 ^ "," ^ string_of_expr e2 ^ ")!=0")
    else string_of_expr e1 ^ string_of_op o ^ string_of_expr e2

  | Sast.Unop (o, e) -> let _,e=e in 
    string_of_uop o ^ string_of_expr e
  | Sast.Access(v,f) ->
        if f="visible" then "v[xx][yy]"
        else if f="type" then "b[xx][yy]"
        else if f="x" then "xx"
        else if f="y" then "yy"
        else failwith ("semantic checker failed, invalid access field")
  | Sast.BoardAccess(x,y) -> 
    let _,x = x and _,y=y in
    "b["^string_of_expr x ^ "][" ^ string_of_expr y ^ "]"
  | Sast.Rand(min,max) ->
      let _,min=min and _,max=max in
      "rand()%("^string_of_expr max^"+1-"^string_of_expr min^")+"^string_of_expr min
  | Sast.Empty -> "\"empty\""




let rec string_of_stmt_list = function
  []      -> ""
  | hd :: tl  ->
    let rec string_of_stmt = function
      | Sast.Expr(expr) -> let _,expr = expr in string_of_expr expr
      | Sast.Return (expr) -> let _,expr = expr in "return " ^ string_of_expr expr ^ ";"
      | Sast.If (e, s1, s2) -> let _,e = e in
          "if(" ^ string_of_expr e ^ ") {" ^ string_of_stmt_list (List.rev s1) ^ "}" ^ (if s2=[] then "" else "else{" ^ string_of_stmt_list (List.rev s2) ^ "}") 
      | Call(f, args)->
                if f = "print" then
                    let _,a0 = List.nth args 0 in 
                    "printf(\"%s\\n\"," ^ string_of_expr a0 ^ ");"
                else if f = "place" then 
                    let _,a0 = List.nth args 0 and _,a1=List.nth args 1 and _,a2= List.nth args 2 in 
                    "addPiece(" ^ string_of_expr a0 ^ "," ^ string_of_expr a1 ^ "," ^ string_of_expr a2
                      ^ ");"
                else if f = "removePiece" then
                    let _,a0 = List.nth args 0 and _,a1=List.nth args 1 in
                      "removePiece(" ^ string_of_expr a0 ^ "," ^ string_of_expr a1 ^ ");"
                else if f="remove" then "removePiece(x,y);"
                else if f="visible" then
                  let _,a0 = List.nth args 0 in
                    "v[x][y]=" ^ string_of_expr a0 ^ ";"
                else if f="setVisibility" then
                  let _,a0 = List.nth args 0 and _,a1 = List.nth args 1 and _,a2 = List.nth args 2 in
                    "v[" ^ string_of_expr a1 ^"]["^string_of_expr a2 ^"]=" ^ string_of_expr a0 ^ ";"

                else if f="click" then
                  let _,a0 = List.nth args 0 and _,a1=List.nth args 1 in
                  "clickSim(" ^ string_of_expr a0 ^ "," ^ string_of_expr a1 ^ ");"
                else if f="changeType" then
                  let _,a0 = List.nth args 0 in
                ( "removePiece(x,y); addPiece(" ^ string_of_expr a0 ^ ",x,y);")
                else if f="endTurn" then ("turnChange = 1;")
                else if f="break" then ("isBroken=1; break;")
                

                else failwith("Undefined function called! Semantic checker failed")
      | Assign (s, e) -> 
        let _,e = e in 
        s ^ "UDV=" ^ string_of_expr e ^ ";"   
      | VDecl (id, e) -> 
        let _,e=e and t,_=e in
        string_of_typ t ^ " " ^ id ^ "UDV = " ^ string_of_expr e ^ ";"
      | ForSurrounding(variable, r, c, s) -> 
          let _,r=r and _,c=c in let r= string_of_expr r and c = string_of_expr c in
          " {int isBroken=0; int atemp, btemp; 
            for(atemp=max(" ^ r ^ "-1,0);atemp<min(WIDTH," ^ r ^ "+2);atemp++){
              for(btemp=max(0," ^ c ^ "-1); btemp<min(HEIGHT," ^ c ^ "+2);btemp++){
                if(isBroken) break;
                int xx = atemp; int yy = btemp;
                  if(!(atemp==" ^ r ^ "&& btemp==" ^ c ^")){ "^
                  string_of_stmt_list (List.rev s)
              ^ "}
                if(isBroken) break;
              }
          }}"
      | ForRow(variable, r, s) -> 
          let _,r=r in let r = string_of_expr r in
          "{int isBroken=0; int ctemp;
          for (ctemp=0;ctemp<WIDTH;ctemp++){ if(isBroken) break;
          int xx=ctemp; int yy=" ^ r ^ "; " ^ 
          string_of_stmt_list (List.rev s)^ "}}" 

      | ForCol(variable, r, s) -> 
          let _,r=r in let r = string_of_expr r in
          "{int isBroken=0; int dtemp;
          for (dtemp=0;dtemp<HEIGHT;dtemp++){ 
          if(isBroken) break;
          int yy=dtemp; int xx=" ^ r ^ "; " ^ 
          string_of_stmt_list (List.rev s)^ "}}" 
      | ForSurroundingCoords (varX, varY, r, c, s) ->  
          let _,r=r and _,c=c in let r= string_of_expr r and c = string_of_expr c in
          " {int isBroken=0; int etemp, ftemp;
            for(etemp=max(" ^ r ^ "-1,0);etemp<min(WIDTH," ^ r ^ "+2);etemp++){
              for(ftemp=max(0," ^ c ^ "-1); ftemp<min(HEIGHT," ^ c ^ "+2);ftemp++){
              if(isBroken) break;
                  if(!(etemp==" ^ r ^ "&& ftemp==" ^ c ^")){
                    int "^varX ^ "UDV=etemp; int "^ varY ^"UDV=ftemp; " ^
                  string_of_stmt_list (List.rev s)
              ^ "}
              if(isBroken) break;
              }
          }}"
      | ForMinMax (variable,min,max, s) ->  
        let _,min=min and _,max=max in let min=string_of_expr min and max = string_of_expr max in
          "{int isBroken=0; int gtemp;
          for (gtemp=" ^ min ^ ";gtemp<=" ^ max ^ ";gtemp++){ if(isBroken) break;
          int " ^ variable ^ "UDV=gtemp;" ^  
          string_of_stmt_list (List.rev s)^ "}}" 
      | ForNested (v1, min1, max1, v2, min2, max2, s) -> 
        let _,min1=min1 and _,max1=max1 and _,min2=min2 and _,max2=max2 in 
        let min1=string_of_expr min1 and max1 = string_of_expr max1 and min2=string_of_expr min2 and max2 = string_of_expr max2 in
          "{int isBroken=0; int htemp, itemp;
          for (htemp=" ^ min1 ^ ";htemp<=" ^ max1 ^ ";htemp++){" ^
              "for (itemp=" ^ min2 ^ ";itemp<=" ^ max2 ^ ";itemp++){ if(isBroken) break;
              int " ^
           v1 ^ "UDV=htemp; int " ^  v2 ^ "UDV=itemp;" ^ 
          string_of_stmt_list (List.rev s)^"}if(isBroken) break;}}" 
      | ForBoard(varX, varY,s) -> 
        "{int isBroken=0; int jtemp, ktemp;
        for (jtemp=0; jtemp<WIDTH;jtemp++){" ^
              "for (ktemp=0;ktemp<HEIGHT;ktemp++){if(isBroken) break;
               int xx=jtemp; int yy=ktemp; int " ^
           varX ^ "UDV=jtemp; int " ^  varY ^ "UDV=ktemp;" ^ 
          string_of_stmt_list (List.rev s)^"}if(isBroken) break;}}" 
      | ForAll (var, s) -> 
          "{ int isBroken=0; int ltemp, mtemp;
          for (ltemp=0; ltemp<WIDTH;ltemp++){" ^
              "for (mtemp=0;mtemp<HEIGHT;mtemp++){ if(isBroken) break;
              int xx=ltemp; int yy=mtemp;" ^ 
          string_of_stmt_list (List.rev s)^"}if(isBroken) break;}}" 

    in (string_of_stmt hd) ^ (string_of_stmt_list tl) 

let string_of_board (board) = 
  let _,x =board.x and _,y=board.y and _,bg=board.img in
  let x=string_of_expr x and y=string_of_expr y and bg = string_of_expr bg 
  in
  "const int WIDTH=" ^ x ^ "; const int HEIGHT=" ^ y ^ "; const char* backgroundImg=" ^ bg ^ "; char * b [" ^ x ^ "][" ^ y ^ "]; int v[" ^x ^ "][" ^ y ^ "];"

let rec string_of_players = function
  [] -> ""
  | hd::[] -> hd
  | hd::tl -> 
      let string_of_player = hd ^ "," 
    in string_of_player ^ string_of_players tl


let playerHeader (players) = "char * players [" ^ (string_of_int (List.length players)) ^ "]={" ^ (string_of_players players) ^ "};"
let turnOrderHeader (turnOrder) = "char* turnOrder [" ^ (string_of_int (List.length turnOrder)) ^ "]={" ^ (string_of_players turnOrder) ^ "};"


let rec piecesHeader = function
  [] -> ""
  |hd::tl ->
      let pieceHeader (piece) =
        let name = piece.name and img = piece.img in
        let name = String.sub name 1 ((String.length name)-2) in
      "const char * " ^ name ^ "IMAGE = " ^ img^ ";" ^ 
      "SDL_Texture * " ^ name ^ "Texture = NULL; void " ^ 
      name ^ "OnClick(int x, int y); void " ^ name ^ "OnTurn(); void " ^ name
      ^ "OnClick(int x, int y){" ^ (string_of_stmt_list (List.rev piece.onClick_typed)) ^ "}" ^ "void "^ name ^ "onTurn(){
        int x,y; for(x=0;x<WIDTH;x++){ for (y=0;y<HEIGHT;y++) { if(strcmp(b[x][y],\"" ^ name ^ "\")==0){" ^ string_of_stmt_list (List.rev piece.onTurn_typed)
            ^ "}}}}"
      in pieceHeader hd ^ piecesHeader tl
let rec piecesInit = function
  [] -> ""
  |hd::tl ->
      let pieceInit (piece) =
        let name = piece.name  in
        let name = String.sub name 1 ((String.length name)-2) in
      name ^ "Texture = loadTexture(" ^ name ^ "IMAGE" ^ ");"
      ^ "if(" ^ name ^ "Texture == NULL ){ printf( \"Failed to load texture image!\\n\" );success = 0;}"
      in pieceInit hd ^ piecesInit tl
let rec piecesClose = function
  [] -> ""
  |hd::tl -> 
      let pieceClose (piece) = 
        let name = piece.name in
        let name = String.sub name 1 ((String.length name)-2) in
      "SDL_DestroyTexture(" ^ name ^ "Texture);" ^ name ^ "Texture=NULL;"
    in pieceClose hd ^ piecesClose tl
let rec piecesToTexture = function
  [] -> "return NULL;"
  |hd::tl -> 
      let pieceToTexture (piece) =  
          let name = piece.name in
          let name = String.sub name 1 ((String.length name)-2) in
          "if(strcmp(piecename, \"" ^ name ^ "\")==0) return " ^ name ^ "Texture; \n else "
    in pieceToTexture hd ^ piecesToTexture tl

let rec callOnTurn = function
  [] -> ""
  |hd::tl ->
      let helper (piece) =
        let name = piece.name  in
        let name = String.sub name 1 ((String.length name)-2) in
        name^"onTurn();"
      in callOnTurn tl ^ helper hd

let textureMatchFunction (pieces) = "SDL_Texture * getTexture(char* piecename){" ^ piecesToTexture pieces ^ "}"


let rec piecesToEvent = function
  [] -> "return;"
  | hd::tl -> 
    let pieceToEvent (piece) = 
      let name = piece.name in
      let name = String.sub name 1 ((String.length name)-2) in
    "if (strcmp (b[x][y], \"" ^ name ^ "\")==0){" ^ name ^ "OnClick(x,y);} else "
  in pieceToEvent hd ^ piecesToEvent tl
let handleEventFunction (pieces) = "void clickSim(int x, int y) {" ^ piecesToEvent pieces ^ "}"


let rec string_of_setup (setup) = string_of_stmt_list setup 

let endFunction (sList, str) = "int " ^ str ^ "(){int atemp, btemp,ctemp, dtemp,etemp,ftemp,gtemp,htemp,itemp,jtemp,ktemp,ltemp,mtemp,isBroken; " ^ (if sList=[] then  "return 0;" else string_of_stmt_list (List.rev sList)) ^ "}"

let string_of_conditions (conditions) = endFunction(conditions.win_typed, "win") ^ endFunction(conditions.lose_typed, "lose") ^ endFunction(conditions.draw_typed, "draw")

(*prints program. currently does not consider anything besides statments. must accomodate our structures*)

let header (board, players, turnOrder, pieces) = "#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
const int SCREEN_WIDTH = 600; const int SCREEN_HEIGHT = 600; const int SCREEN_BPP = 32; const int DISPLAY_HEIGHT = 000; 
int init();
int loadMedia();
void closeSDL();
SDL_Texture * getTexture (char * piecename);
void renderPiece(char * piecename, int x, int y);
int addPiece(char * piecename, int x, int y);
int removePiece(int x, int y);
SDL_Texture * loadTexture (char * filename);
void handleEvent( SDL_Event* e );
int xCord(int x);
int yCord(int y);
void clickSim(int x, int y);
int win();
int lose();
int draw();
int min();
int max();
SDL_Surface * backgroundTex = NULL;
SDL_Window* gWindow = NULL;
SDL_Surface* gRenderer = NULL;
int turnChange = 0;
char * turn;" ^(string_of_board board) ^ (piecesHeader pieces) ^ (playerHeader players) ^ (turnOrderHeader turnOrder) 

let staticFunctions = "
void handleEvent( SDL_Event* e )
{
    //Get mouse position
    int x, y;
    SDL_GetMouseState( &x, &y );
    x=xCord(x); y=yCord(y);
   clickSim(x,y);
}
void renderPiece(char * piecename, int x, int y){
  SDL_Rect myRect = { x * SCREEN_WIDTH/WIDTH, y * SCREEN_HEIGHT/HEIGHT, SCREEN_WIDTH/WIDTH, SCREEN_HEIGHT/HEIGHT};
  SDL_RenderCopy( gRenderer, getTexture(piecename), NULL,  &(myRect) );
}
int addPiece(char * piecename, int x, int y){
  if(strcmp(b[x][y],\"empty\")!=0){
      return 0;
  }
  b[x][y] = piecename;
  v[x][y] = 1;
  return 1;
}
int removePiece (int x, int y){
  if (strcmp(b[x][y],\"empty\")==0){
    return 0;
  }
  b[x][y] = \"empty\";
  v[x][y] = 0;
  return 1;
}
SDL_Texture* loadTexture( char * filename )
{
  SDL_Texture* newTexture = NULL;
  SDL_Surface* loadedSurface = IMG_Load( filename );
  if( loadedSurface == NULL )
  {
    printf( \"Unable to load image %s! SDL_image Error: %s\\n\", filename, IMG_GetError() );
  }
  else
  {
        newTexture = SDL_CreateTextureFromSurface( gRenderer, loadedSurface );
    if( newTexture == NULL )
    {
      printf( \"Unable to create texture from %s! SDL Error: %s\\n\", filename, SDL_GetError() );
    }
    SDL_FreeSurface( loadedSurface );
  }
  return newTexture;
}


int init()
{
  int success = 1;
  turn = turnOrder[0];

  int i,j;
  for(i = 0; i< sizeof(b) / sizeof((b)[0]) ; i++){
    for(j=0; j < sizeof(b[i])/sizeof(b[i][0]); j++){
      b[i][j]=\"empty\";
      v[i][j] = 0;
    }
  }

  if( SDL_Init( SDL_INIT_VIDEO ) < 0 )
  {
    printf( \"SDL could not initialize! SDL Error: %s\\n\", SDL_GetError() );
    success = 0;
  }
  else
  {
    if( !SDL_SetHint( SDL_HINT_RENDER_SCALE_QUALITY, \"1\" ) )
    {
      printf( \"Warning: Linear texture filtering not enabled!\" );
    }
    gWindow = SDL_CreateWindow( \"Game\", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, SCREEN_WIDTH, SCREEN_HEIGHT + DISPLAY_HEIGHT, SDL_WINDOW_SHOWN );
    if( gWindow == NULL )
    {
      printf( \"Window could not be created! SDL Error: %s\\n\", SDL_GetError() );
      success = 0;
    }
    else
    {
      gRenderer = SDL_CreateRenderer( gWindow, -1, SDL_RENDERER_ACCELERATED );
      if( gRenderer == NULL )
      {
        printf( \"Renderer could not be created! SDL Error: %s\\n\", SDL_GetError() );
        success = 0;
      }
      else
      {
        SDL_SetRenderDrawColor( gRenderer, 0xFF, 0xFF, 0xFF, 0xFF );
        int imgFlags = IMG_INIT_PNG;
        if( !( IMG_Init( imgFlags ) & imgFlags ) )
        {
          printf( \"SDL_image could not initialize! SDL_image Error: %s\\n\", IMG_GetError() );
          success = 0;
        }
      }
    }
  }

  return success;
}

int xCord(int x){
  int i;
  for(i=0;i<WIDTH;i++){
    if(x>i && x<(i+1)*SCREEN_WIDTH/WIDTH){
      return i;
    }
  }

  return -1;
}

int yCord(int y){
  int i;
  for(i =0; i<HEIGHT; i++){
    if(y>i && y<(i+1)*SCREEN_HEIGHT/HEIGHT){
      return i;
    }
  }  
  return -1;
}

int min(int a, int b){
  if (a<b)
    return a;
  return b;
}

int max(int a, int b){
  if (a>b)
    return a;
  return b;
}

"

let loadAndCloseFunctions (pieces)= "int loadMedia(){int success = 1;backgroundTex = loadTexture( backgroundImg );if( backgroundTex == NULL ){printf( \"Failed to load texture image!\\n\" );success = 0;
}" ^ piecesInit pieces^"return success;}" ^ "void closeSDL(){SDL_DestroyTexture( backgroundTex );backgroundTex = NULL;" ^ piecesClose pieces ^ "SDL_DestroyRenderer( gRenderer ); SDL_DestroyWindow( gWindow );
gWindow = NULL;gRenderer = NULL; IMG_Quit();SDL_Quit();}"


let mainFunction (pieces, setup)= "int main( int argc, char* args[] )
{
int atemp, btemp,ctemp, dtemp,etemp,ftemp,gtemp,htemp,itemp,jtemp,ktemp,ltemp,mtemp,  isBroken;
  if( !init() )
  {
    printf( \"Failed to initialize!\\n\" );
  }
  else
  {   if( !loadMedia() )
    {
      printf( \"Failed to load media!\\n\" );
    }
    else
    { 
      int quit = 0;" ^ string_of_setup setup ^ " SDL_Event e;            
      int turnCounter = 0;


  while( !quit && !win() && !lose() && !draw())
      {
        while( SDL_PollEvent( &e ) != 0 )
        {
          if( e.type == SDL_QUIT )
          {
            quit = 1;
          }
          else if( e.type == SDL_MOUSEBUTTONDOWN /*|| e->type == SDL_MOUSEBUTTONUP*/ ){
              handleEvent(&e);
          }
        }
        if(turnChange){
          turnChange = 0;
          turnCounter = (turnCounter + 1) % ( sizeof(turnOrder) / sizeof(turnOrder[0]));
          turn = turnOrder[turnCounter];"
            ^ callOnTurn pieces^
        "}
        SDL_SetRenderDrawColor( gRenderer, 0xFF, 0xFF, 0xFF, 0xFF );
        SDL_RenderClear( gRenderer );

        SDL_Rect board;
        board.x = 0; board.y = 0;
        board.w = SCREEN_WIDTH; board.h=SCREEN_HEIGHT;
        SDL_RenderSetViewport(gRenderer, &board);

        SDL_RenderCopy( gRenderer, backgroundTex, NULL, NULL );
        int i, j;
        for(i=0;i<=HEIGHT;i++)
        {
          //y location at i*screenheight/height
          int y = i * SCREEN_HEIGHT/HEIGHT;
          if (y==SCREEN_HEIGHT)
            y-=1;
          SDL_SetRenderDrawColor( gRenderer, 0x00, 0x00, 0x00, 0xFF );    
          SDL_RenderDrawLine( gRenderer, 0, y, SCREEN_WIDTH, y );

        }
        for(i=0;i<=WIDTH;i++)
        {

          int x = i * SCREEN_WIDTH/WIDTH;
          if (x==SCREEN_WIDTH)
            x-=1;
          SDL_SetRenderDrawColor( gRenderer, 0x00, 0x00, 0x00, 0xFF );    
          SDL_RenderDrawLine( gRenderer, x, 0, x, SCREEN_HEIGHT );
        }
        for(i = 0; i< sizeof(b) / sizeof((b)[0]) ; i++){
          for(j=0; j < sizeof(b[i])/sizeof(b[i][0]); j++){
            if(strcmp(b[i][j],\"empty\")!=0 && v[i][j]==1){
              renderPiece(b[i][j],i,j);
            }
          }
        }
        SDL_Rect messageDisplay;
        messageDisplay.x=SCREEN_HEIGHT; messageDisplay.y = 0;
        messageDisplay.w = SCREEN_WIDTH; messageDisplay.h = DISPLAY_HEIGHT;
        SDL_RenderSetViewport (gRenderer, &messageDisplay);



        SDL_RenderPresent( gRenderer );



      }
    }
  }
  closeSDL();

  return 0;
}"


let printprog (board, players, turnOrder, pieces, setup, conditions) = 
  let oc = open_out file in 
      fprintf oc "%s" (header (board,players, turnOrder, pieces));
            fprintf oc "%s" (textureMatchFunction pieces);
            fprintf oc "%s" staticFunctions;
            fprintf oc "%s" (loadAndCloseFunctions pieces);
            fprintf oc "%s" (handleEventFunction pieces);
            fprintf oc "%s" (string_of_conditions conditions);
            fprintf oc "%s" (mainFunction (pieces,(List.rev setup)));

  close_out oc;

